"""The :py:mod:`in_toolset.model` module contains code for modelling the behaviour of petri nets, industry nets and generic uncoloured petri nets.

It relies heavily on the signalling and property classes in :py:mod:`in_toolset.common`."""
